#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import math
import numpy as np

import separation_base
import audio_signal
import constants


class Nmf(separation_base.SeparationBase):
    """
    This is an implementation of the Non-negative Matrix Factorization algorithm for
    source separation. This implementation cannot receive a raw audio signal, rather
    it only accepts a stft and a number, num_templates, which defines the number of bases
    vectors.

    This class provides two implementations of distance measures, EUCLIDEAN and DIVERGENCE,
    and also allows the user to define distance measure function.

    References:
    [1] Lee, Daniel D., and H. Sebastian Seung. "Algorithms for non-negative matrix factorization."
        Advances in neural information processing systems. 2001.
    """

    def __str__(self):
        return "Nmf"

    # TODO: Change this so that NMF accepts an AudioSignal object and not a raw stft
    def __init__(self, stft, num_templates, input_audio_signal=None, sample_rate=None, stft_params=None,
                 activation_matrix=None, templates=None, distance_measure=None,
                 should_update_template=None, should_update_activation=None):
        self.__dict__.update(locals())
        super(Nmf, self).__init__(input_audio_signal=input_audio_signal,
                                  sample_rate=sample_rate, stft_params=stft_params)

        if num_templates <= 0:
            raise Exception('Need more than 0 bases!')

        if stft.size <= 0:
            raise Exception('STFT size must be > 0!')

        self.stft = stft  # V, in literature
        self.num_templates = num_templates

        if activation_matrix is None and templates is None:
            self.templates = np.zeros((stft.shape[0], num_templates))  # W, in literature
            self.activation_matrix = np.zeros((num_templates, stft.shape[1]))  # H, in literature
            self.randomize_input_matrices()
        elif activation_matrix is not None and templates is not None:
            self.templates = templates
            self.activation_matrix = activation_matrix
        else:
            raise Exception('Must provide both activation matrix and template vectors or nothing at all!')

        self.distance_measure = distance_measure if distance_measure is not None else DistanceType.DEFAULT
        self.should_update_template = True if should_update_template is None else should_update_template
        self.should_update_activation = True if should_update_activation is None else should_update_activation

        self.should_use_epsilon = False  # Replace this with something more general
        self.epsilon_euclidean_type = True
        self.stopping_epsilon = 1e10
        self.max_num_iterations = 20

    def run(self):
        """
        This runs the NMF separation algorithm. This function assumes that all
        parameters have been set prior to running.

        No inputs. do_STFT and N must be set prior to calling this function.

        Returns an activation matrix (in a 2d numpy array)
        and a set of template vectors (also 2d numpy array).
        """

        if self.stft is None or self.stft.size == 0:
            raise Exception('Cannot do NMF with an empty STFT!')

        if self.num_templates is None or self.num_templates == 0:
            raise Exception('Cannot do NMF with no bases!')

        if self.should_use_epsilon:
            print 'Warning: User is expected to have set stopping_epsilon prior to using' \
                  ' this function. Expect this to take a long time if you have not set' \
                  ' a suitable epsilon.'

        should_stop = False
        num_iterations = 0
        while not should_stop:

            self.update()

            # Stopping conditions
            num_iterations += 1
            if self.should_use_epsilon:
                if self.epsilon_euclidean_type:
                    should_stop = self._euclidean_distance() <= self.stopping_epsilon
                else:
                    should_stop = self._divergence() <= self.stopping_epsilon
            else:
                should_stop = num_iterations >= self.max_num_iterations

        return self.activation_matrix, self.templates

    def update(self):
        """
        Computes a single update using the update function specified.
        :return: nothing
        """
        # update activation matrix
        if self.should_update_activation:
            if self.distance_measure == DistanceType.EUCLIDEAN:
                self.activation_matrix = self._update_activation_euclidean()

            elif self.distance_measure == DistanceType.DIVERGENCE:
                self.activation_matrix = self._update_activation_divergent()

        # update template vectors
        if self.should_update_template:
            if self.distance_measure == DistanceType.EUCLIDEAN:
                self.templates = self._update_template_euclidean()

            elif self.distance_measure == DistanceType.DIVERGENCE:
                self.templates = self._update_template_divergence()

    def _update_activation_euclidean(self):
        # make a new matrix to store results
        activation_copy = np.empty_like(self.activation_matrix)

        # store in memory so we don't have to do n*m calculations.
        template_T = self.templates.T
        temp_T_stft = np.dot(template_T, self.stft)
        temp_T_act = np.dot(np.dot(template_T, self.templates), self.activation_matrix)

        # Eq. 4, H update from [1]
        for indices, val in np.ndenumerate(self.activation_matrix):
            result = temp_T_stft[indices]
            result /= temp_T_act[indices]
            result *= self.activation_matrix[indices]
            activation_copy[indices] = result

        return activation_copy

    def _update_template_euclidean(self):
        # make a new matrix to store results
        template_copy = np.empty_like(self.templates)

        # store in memory so we don't have to do n*m calculations.
        activation_T = self.activation_matrix.T
        stft_act_T = np.dot(self.stft, activation_T)
        temp_act = np.dot(np.dot(self.templates, self.activation_matrix), activation_T)

        # Eq. 4, W update from [1]
        for indices, val in np.ndenumerate(self.templates):
            result = stft_act_T[indices]
            result /= temp_act[indices]
            result *= self.templates[indices]
            template_copy[indices] = result

        return template_copy

    def _update_activation_divergent(self):
        # make a new matrix to store results
        activation_copy = np.empty_like(self.activation_matrix)

        dot = np.dot(self.templates, self.activation_matrix)

        # Eq. 5, H update from [1]
        for indices, val in np.ndenumerate(self.activation_matrix):
            (a, mu) = indices
            result = sum((self.templates[i][a] * self.stft[i][mu]) / dot[i][mu]
                         for i in range(self.templates.shape[0]))
            result /= sum(self.templates[k][a] for k in range(self.templates.shape[0]))
            result *= self.activation_matrix[indices]
            activation_copy[indices] = result

        return activation_copy

    def _update_template_divergence(self):
        # make a new matrix to store results
        template_copy = np.empty_like(self.templates)

        dot = np.dot(self.templates, self.activation_matrix)

        # Eq. 5, W update from [1]
        for indices, val in np.ndenumerate(self.templates):
            (i, a) = indices
            result = sum((self.activation_matrix[a][mu] * self.stft[i][mu]) / dot[i][mu]
                         for mu in range(self.activation_matrix.shape[1]))
            result /= sum(self.activation_matrix[a][nu] for nu in range(self.activation_matrix.shape[1]))
            result *= self.templates[indices]
            template_copy[indices] = result

        return template_copy

    def _euclidean_distance(self):
        try:
            mixture = np.dot(self.templates, self.activation_matrix)
        except:
            print self.activation_matrix.shape, self.templates.shape
            return

        if mixture.shape != self.stft.shape:
            raise Exception('Something went wrong! Recombining the activation matrix '
                            'and template vectors is not the same size as the STFT!')

        return sum((self.stft[index] - val) ** 2 for index, val in np.ndenumerate(mixture))

    def _divergence(self):
        mixture = np.dot(self.activation_matrix, self.templates)

        if mixture.shape != self.stft.shape:
            raise Exception('Something went wrong! Recombining the activation matrix '
                            'and template vectors is not the same size as the STFT!')

        return sum(
            (self.stft[index] * math.log(self.stft[index] / val, 10) + self.stft[index] - val)
            for index, val in np.ndenumerate(mixture))

    def make_audio_signals(self):
        # TODO: this!
        raise NotImplementedError('This does not work yet.')
        signals = []
        for stft in self.recombine_calculated_matrices():
            signal = audio_signal.AudioSignal(stft=stft)
            signal.istft()
            signals.append(signal)
        return signals

    def recombine_calculated_matrices(self):
        new_matrices = []
        for n in range(self.num_templates):
            matrix = np.empty_like(self.activation_matrix)
            matrix[n, ] = self.activation_matrix[n, ]

            new_stft = np.dot(self.templates, matrix)
            new_matrices.append(new_stft)
        return new_matrices

    def randomize_input_matrices(self, shouldNormalize=False):
        self._randomize_matrix(self.activation_matrix, shouldNormalize)
        self._randomize_matrix(self.templates, shouldNormalize)

    @staticmethod
    def _randomize_matrix(M, shouldNormalize=False):
        for i, row in enumerate(M):
            for j, col in enumerate(row):
                M[i][j] = random.random()

                if not shouldNormalize:
                    M[i][j] *= constants.DEFAULT_MAX_VAL
        return M

    def plot(self, outputFile, **kwargs):
        raise NotImplementedError('Sorry, you cannot do this yet.')


class DistanceType:
    EUCLIDEAN = 'euclidean'
    DIVERGENCE = 'divergence'
    DEFAULT = EUCLIDEAN

    def __init__(self):
        pass



import time
import numpy as np
import nussl

def simple_example():
    """
    A simple example of NMF in nussl
    """
    # Make two simple matrices
    n = 4
    a = np.arange(n ** 2).reshape((n, n))
    b = 2 * a + 3

    # Mix them together
    mixture = np.dot(b, a)

    # Set up NU NMF
    nBases = 2
    nmf = nussl.Nmf(mixture, nBases)
    nmf.should_use_epsilon = False
    nmf.max_num_iterations = 100
    nmf.distance_measure = nussl.DistanceType.EUCLIDEAN

    # run NU NMF
    start = time.time()
    nmf.run()
    print '{0:.3f}'.format(time.time() - start), 'seconds for NUSSL'

    print 'Original mixture =\n', mixture
    print 'NMF Reconstructed mixture =\n', np.dot(nmf.templates, nmf.activation_matrix)

    print 'Sources ='
    signals = nmf.recombine_calculated_matrices()
    for sig in signals:
        print sig


if __name__ == '__main__':
    simple_example()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import spectral_utils
import constants
import audio_signal


class SeparationBase(object):
    """Base class for all separation algorithms.

    Do not call this. It will not do anything.

    self.audio_signal, self.sample_rate, and self.stft_params are all properties that can have their setters and getters
    overridden in subclasses.

    Parameters:
        audio_signal (Optional[np.array]): Audio signal in array form. Defaults to AudioSignal.AudioSignal()
        sample_rate (Optional[int]): Sample rate. Defaults to Constants.DEFAULT_SAMPLE_RATE
        stft_params (StftParams): STFT parameters for the separation algorithm. Defaults to
         spectral_utils.StftParams(Constants.DEFAULT_SAMPLE_RATE)
    """

    def __init__(self, input_audio_signal=None, sample_rate=None, stft_params=None):

        if input_audio_signal is not None:
            self.audio_signal = input_audio_signal
        else:
            self.audio_signal = audio_signal.AudioSignal()

        if sample_rate is not None:
            self.sample_rate = sample_rate
        else:
            self.sample_rate = constants.DEFAULT_SAMPLE_RATE

        if stft_params is not None:
            self.stft_params = stft_params
        else:
            self.stft_params = spectral_utils.StftParams(self.sample_rate)


    def plot(self, output_name, **kwargs):
        """Plots relevant data for separation algorithm

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def run(self):
        """run separation algorithm

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def make_audio_signals(self):
        """Makes AudioSignal objects after separation algorithm is run

        Raises:
            NotImplementedError: Cannot call base class
        """
        raise NotImplementedError('Cannot call base class.')

    def __call__(self):
        self.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import scipy.fftpack as scifft
from scipy.signal import hamming, hann, blackman
import os.path
import librosa

import constants


def plot_stft(signal, file_name, title=None, win_length=None, hop_length=None,
              window_type=None, sample_rate=None, n_fft_bins=None,
              freq_max=None, show_interactive_plot=False):
    """
    Outputs an image of an stft plot of input audio, signal. This uses matplotlib to create the output file.
    You can specify the same all of the same parameters that are in e_stft(). By default, the StftParams defaults
    are used for any values not provided in (win_length, hop_length, and window_type).
    Title is settable by user and there is a flag to show an interactive matplotlib graph, as well.

    Notes:
        To find out what output formats are available for your machine run the following code:
        ::

            import matplotlib.pyplot as plt
            fig = plt.figure()

            print fig.canvas.get_supported_filetypes()

        (From here: http://stackoverflow.com/a/7608273/5768001)

    Args:
        signal: (np.array) input time series signal that will be plotted
        file_name: (str) path to file that will be output. Will overwrite any file that is already there.
        Uses mat
        title: (string) (Optional) Title to go at top of graph. Defaults to 'Spectrogram of [file_name]'
        win_length: (int) (Optional) number of samples per window. Defaults to StftParams default.
        hop_length: (int) (Optional) number of samples between the start of adjacent windows, or "hop".
        Defaults to StftParams default.
        sample_rate: (int) (Optional) sample rate of input signal.  Defaults to StftParams default.
        window_type: (string) (Optional) type of window to use. Using WindowType object is recommended.
        Defaults to StftParams default.
        n_fft_bins: (int) (Optional) number of fft bins per time window.
        If not specified, defaults to next highest power of 2 above window_length. Defaults to StftParams default.
        freq_max: (int) Max frequency to display. Defaults to 44100Hz
        show_interactive_plot: (bool) (Optional) Flag indicating if plot should be shown when function is run.
        Defaults to False

    Examples::

        # Set up sine wave parameters
        sr = nussl.Constants.DEFAULT_SAMPLE_RATE # 44.1kHz
        n_sec = 3 # seconds
        duration = n_sec * sr
        freq = 300 # Hz

        # Make sine wave array
        x = np.linspace(0, freq * 2 * np.pi, duration)
        x = np.sin(x)

        # plot it and save it in path 'path/to/sine_wav.png'
        nussl.plot_stft(x, 'path/to/sine_wav.png')

    """
    sample_rate = constants.DEFAULT_SAMPLE_RATE if sample_rate is None else sample_rate
    freq_max = constants.MAX_FREQUENCY if freq_max is None else freq_max

    if title is None:
        title = os.path.basename(file_name)
        title = os.path.splitext(title)[0]
        title = 'Spectrogram of {}'.format(title)

    required = [win_length, hop_length, window_type]
    if any([r is None for r in required]):
        defaults = StftParams(sample_rate)

        win_length = defaults.window_length if win_length is None else win_length
        hop_length = defaults.hop_length if hop_length is None else hop_length
        window_type = defaults.window_type if window_type is None else window_type

    (stft, psd, freqs, time) = e_stft_plus(signal, win_length, hop_length, window_type, sample_rate, n_fft_bins)

    plt.close('all')

    # TODO: remove transposes!
    time_tile = np.tile(time, (len(freqs), 1))
    freq_tile = np.tile(freqs.T, (len(time), 1)).T
    sp = librosa.logamplitude(np.abs(stft) ** 2, ref_power=np.max)
    plt.pcolormesh(time_tile, freq_tile, sp)

    plt.axis('tight')
    plt.xlabel('Time (sec)')
    plt.ylabel('Frequency (Hz)')
    plt.title(title)
    plt.ylim(freqs[0], freq_max)

    plt.savefig(file_name)

    if show_interactive_plot:
        plt.interactive('True')
        plt.show()


def e_stft(signal, window_length, hop_length, window_type, n_fft_bins=None, remove_reflection=True, use_librosa=False):
    """
    This function computes a short time fourier transform (STFT) of a 1D numpy array input signal.
    This will zero pad the signal by half a hop_length at the beginning to reduce the window
    tapering effect from the first window. It also will zero pad at the end to get an integer number of hops.

    By default, this function removes the FFT data that is a reflection from over Nyquist. There is an option
    to suppress this behavior and have this function include data from above Nyquist, but since the
    inverse STFT function, e_istft(), expects data without the reflection, the onus is on the user to remember
    to set the reconstruct_reflection flag in e_istft() input.

    Additionally, this function assumes a single channeled audio signal and is not guaranteed to work on
    multichannel audio. If you want to do an STFT on multichannel audio see the AudioSignal object.

    Args:
        signal: 1D numpy array containing audio data. (REAL?COMPLEX?INTEGER?)
        window_length: (int) number of samples per window
        hop_length: (int) number of samples between the start of adjacent windows, or "hop"
        window_type: (string) type of window to use. Using WindowType object is recommended.
        n_fft_bins: (int) (Optional) number of fft bins per time window.
        If not specified, defaults to next highest power of 2 above window_length
        remove_reflection: (bool) (Optional) if True, this will remove reflected STFT data above the Nyquist point.
        If not specified, defaults to True.
        use_librosa: (bool) (Optional) This flag bypasses nussl's stft function and will call librosa's stft. nussl
        will massage the output so that it is in a format that it expects. remove_reflection is still works in this
        mode. Note: librosa's works differently than nussl's and may produce different output.

    Returns:
        2D  numpy array with complex STFT data.
        Data is of shape (num_time_blocks, num_fft_bins). These numbers are determined by length of the input signal,
        on internal zero padding (explained at top), and n_fft_bins/remove_reflection input (see example below).

    Examples::

        # Set up sine wave parameters
        sr = nussl.Constants.DEFAULT_SAMPLE_RATE # 44.1kHz
        n_sec = 3 # seconds
        duration = n_sec * sr
        freq = 300 # Hz

        # Make sine wave array
        x = np.linspace(0, freq * 2 * np.pi, duration)
        x = np.sin(x)

        # Set up e_stft() parameters
        win_type = nussl.WindowType.HANN
        win_length = 2048
        hop_length = win_length / 2

        # Run e_stft()
        stft = nussl.e_stft(x, win_length, hop_length, win_type)
        # stft has shape (win_length // 2 + 1 , duration / hop_length)

        # Get reflection
        stft_with_reflection = nussl.e_stft(x, win_length, hop_length, win_type, remove_reflection=False)
        # stft_with_reflection has shape (win_length, duration / hop_length)

        # Change number of fft bins per hop
        num_bins = 4096
        stft_more_bins = e_stft(x, win_length, hop_length, win_type, n_fft_bins=num_bins)
        # stft_more_bins has shape (num_bins // 2 + 1, duration / hop_length)

    """
    if n_fft_bins is None:
        n_fft_bins = int(2 ** np.ceil(np.log2(window_length)))

    if use_librosa:
        stft = librosa.stft(signal, n_fft_bins, hop_length, window_length,
                            make_window(window_type, window_length))

        # librosa removes the reflection by default, so we have to reconstruct it in this case
        if not remove_reflection:
            stft = _add_reflection(stft)

        return stft

    orig_signal_length = len(signal)

    # zero-pad the vector at the beginning and end to reduce the window tapering effect
    if window_length % 2 == 0:
        zero_pad1 = np.zeros(window_length / 2)
    else:
        zero_pad1 = np.zeros((window_length - 1) / 2)
    signal = np.concatenate((zero_pad1, signal, zero_pad1))

    # another zero pad if not integer multiple of hop_length
    zero_pad2_len = 0
    if orig_signal_length % hop_length != 0:
        zero_pad2_len = hop_length - (orig_signal_length % hop_length)
        signal = np.concatenate((signal, np.zeros(zero_pad2_len)))

    window = make_window(window_type, window_length)

    # figure out size of output stft
    num_blocks = int(((len(signal) - window_length) / hop_length + 1))
    stft_bins = n_fft_bins / 2 + 1 if remove_reflection else n_fft_bins  # only want just over half of each fft

    # this is where we do the stft calculation
    stft = np.zeros((num_blocks, stft_bins), dtype=complex)
    for hop in range(num_blocks):
        start = hop * hop_length
        end = start + window_length
        unwindowed_signal = signal[start:end]
        windowed_signal = np.multiply(unwindowed_signal, window)
        fft = scifft.fft(windowed_signal, n=n_fft_bins)
        stft[hop, ] = fft[0:stft_bins]

    # reshape the 2d array, so it's how we expect it.
    stft = stft.T
    if remove_reflection:
        first = int(len(zero_pad1) / hop_length)
        last = stft.shape[1] - int((len(zero_pad1) + zero_pad2_len) / hop_length)
        stft = stft[:, first: last]

    return stft


def e_istft(stft, window_length, hop_length, window_type, reconstruct_reflection=True, remove_padding=False,
            use_librosa=False):
    """
    Computes an inverse short time fourier transform (STFT) from a 2D numpy array of complex values. By default
    this function assumes input STFT has no reflection above Nyquist and will rebuild it, but the
    reconstruct_reflection flag overrides that behavior.

    Additionally, this function assumes a single channeled audio signal and is not guaranteed to work on
    multichannel audio. If you want to do an iSTFT on multichannel audio see the AudioSignal object.

    Args:
        stft: complex valued 2D numpy array containing STFT data
        window_length: (int) number of samples per window
        hop_length: (int) number of samples between the start of adjacent windows, or "hop"
        window_type: (deprecated)
        reconstruct_reflection: (bool) (Optional) if True, this will recreate the removed reflection
        data above the Nyquist. If False, this assumes that the input STFT is complete. Default is True.
        remove_padding: (bool) (Optional) if True, this function will remove the first and
            last (window_length - hop_length) number of samples. Defaults to False.
        use_librosa: (bool) (Optional) This flag bypasses nussl's istft function and will call librosa's istft. nussl
        will massage the output so that it is in a format that it expects. remove_reflection is still works in this
        mode. Note: librosa's works differently than nussl's and may produce different output.

    Returns:
        1D numpy array containing an audio signal representing the original signal used to make stft

    Examples::

        # Set up sine wave parameters
        sr = nussl.Constants.DEFAULT_SAMPLE_RATE # 44.1kHz
        n_sec = 3 # seconds
        duration = n_sec * sr
        freq = 300 # Hz

        # Make sine wave array
        x = np.linspace(0, freq * 2 * np.pi, duration)
        x = np.sin(x)

        # Set up e_stft() parameters
        win_type = nussl.WindowType.HANN
        win_length = 2048
        hop_length = win_length / 2

        # Get an stft
        stft = nussl.e_stft(x, win_length, hop_length, win_type)

        calculated_signal = nussl.e_istft(stft, win_length, hop_length)

    """
    if use_librosa:
        signal = librosa.istft(stft, hop_length, window_length, make_window(window_type, window_length))
        return signal

    n_hops = stft.shape[1]
    signal_length = (n_hops - 1) * hop_length + window_length
    signal = np.zeros(signal_length)

    # Add reflection back
    if reconstruct_reflection:
        stft = _add_reflection(stft)

    for n in range(n_hops):
        start = n * hop_length
        end = start + window_length
        signal[start:end] = signal[start:end] + np.real(scifft.ifft(stft[:, n]))

    # remove zero-padding
    if remove_padding:
        start = window_length - hop_length
        stop = signal_length - (window_length - hop_length)
        signal = signal[start:stop]

    return signal


def e_stft_plus(signal, window_length, hop_length, window_type, sample_rate, n_fft_bins=None):
    """
    Does a short time fourier transform (STFT) of the signal (by calling e_stft() ), but also calculates
    the power spectral density (PSD), frequency and time vectors for the calculated STFT. This function does not
    give you as many options as e_stft() (wrt removing the reflection and using librosa). If you need that
    flexibility, it is recommended that you either use e_stft() or use an AudioSignal object.

    Use this is situations where you need more than just the STFT data. For instance, this is used in plot_stft()
    to get the frequency vector to graph. In situations where you don't need this extra data it is
    more efficient to use e_stft().

    Additionally, this function assumes a single channeled audio signal and is not guaranteed to work on
    multichannel audio. If you want to do an STFT on multichannel audio see the AudioSignal object.

    Args:
        signal: 1D numpy array containing audio data. (REAL?COMPLEX?INTEGER?)
        window_length: (int) number of samples per window
        hop_length: (int) number of samples between the start of adjacent windows, or "hop"
        window_type: (string) type of window to use. Using WindowType object is recommended.
        sample_rate: (int) the intended sample rate, this is used in the calculation of the frequency vector
        n_fft_bins: (int) (Optional) number of fft bins per time window.
        If not specified, defaults to next highest power of 2 above window_length

    Returns:
        stft: (np.ndarray) a 2D matrix short time fourier transform data

    """

    if n_fft_bins is None:
        n_fft_bins = window_length

    stft = e_stft(signal, window_length, hop_length, window_type, n_fft_bins)
    frequency_vector = (sample_rate / 2) * np.linspace(0, 1, (n_fft_bins / 2) + 1)

    time_vector = np.array(range(stft.shape[1]))
    hop_in_secs = hop_length / (1.0 * sample_rate)
    time_vector = np.multiply(hop_in_secs, time_vector)

    window = make_window(window_type, window_length)
    win_dot = np.dot(window, window.T)
    psd = np.zeros_like(stft, dtype=float)
    for i in range(psd.shape[1]):
        psd[:, i] = (1 / float(sample_rate)) * ((abs(stft[:, i]) ** 2) / float(win_dot))

    return stft, psd, frequency_vector, time_vector


def make_window(window_type, length):
    """Returns an np array of type window_type

    Args:
        window_type (WindowType): Type of window to create, window_type object
        length (int): length of window

    Returns:
        window (np.array): np array with a window of type window_type
    """

    # Generate samples of a normalized window
    if window_type == WindowType.RECTANGULAR:
        return np.ones(length)
    elif window_type == WindowType.HANN:
        return hann(length, False)
    elif window_type == WindowType.BLACKMAN:
        return blackman(length, False)
    elif window_type == WindowType.HAMMING:
        # return np.hamming(length)
        return hamming(length, False)
    else:
        return None


def _add_reflection(matrix):
    reflection = matrix[-2:0:-1, :]
    reflection = reflection.conj()
    return np.vstack((matrix, reflection))


class WindowType:
    """
    The WindowType class provides standardized strings for use in make_window(). Many other things
    in spectral_utils.py use this class, but they all get bubbled down to make_window().
    The windows defined are:
        * ``RECTANGULAR = 'rectangular'``
        * ``HAMMING = 'hamming'``
        * ``HANN = 'hann'``
        * ``BLACKMAN = 'blackman'``
        * ``DEFAULT = HAMMING``

    Examples::

        # get a blackman window
        window_type = nussl.WindowType.BLACKMAN
        length = 1024
        blackman = nussl.make_window(window_type, length)

        # Load audio file and take stft with rectangular window
        audio = nussl.AudioSignal('path/to/signal.wav')
        audio.stft(window_type = WindowType.RECTANGULAR) # this will only set 'rectangular' window for this call

        # But we can make it so every time anything does an STFT
        # on the audio object it uses a rectangular window like this:
        audio.StftParams.window_type = WindowType.RECTANGULAR

        my_stft = audio.stft() # uses rectangular window because we set it in previous line

    """
    RECTANGULAR = 'rectangular'
    HAMMING = 'hamming'
    HANN = 'hann'
    BLACKMAN = 'blackman'
    DEFAULT = HAMMING

    all_types = [RECTANGULAR, HAMMING, HANN, BLACKMAN]

    def __init__(self):
        pass


class StftParams(object):
    """
    The StftParams class is a container for information needed to run an STFT or iSTFT.
    This is meant as a convenience and does not actually perform any calculations within. It should
    get "decomposed" by the time e_stft() or e_istft() are called, so that every attribute in this
    object is a parameter to one of those functions.

    Every class that inherits from the SeparationBase class has an StftParms object, and this
    is the only way that a top level user has access to the STFT parameter settings that
    all of the separation algorithms are built upon.
    This object will get passed around instead of each of these individual attributes.
    """
    def __init__(self, sample_rate, window_length=None, hop_length=None, window_type=None, n_fft_bins=None):
        self.sample_rate = sample_rate

        # default to 40ms windows
        default_win_len = int(2 ** (np.ceil(np.log2(constants.DEFAULT_WIN_LEN_PARAM * sample_rate))))
        self._window_length = default_win_len if window_length is None else window_length
        self._hop_length = self._window_length / 2 if hop_length is None else hop_length
        self.window_type = WindowType.DEFAULT if window_type is None else window_type
        self._n_fft_bins = self._window_length if n_fft_bins is None else n_fft_bins

        self._hop_length_needs_update = True
        self._n_fft_bins_needs_update = True

        if hop_length is not None:
            self._hop_length_needs_update = False

        if n_fft_bins is not None:
            self._n_fft_bins_needs_update = False

    @property
    def window_length(self):
        """
        Length of stft window in samples. If window_overlap
        or num_fft are not set manually, then changing this will update them to
        ``hop_length = window_length // 2``, and and ``num_fft = window_length``
        **This property is settable.**
        """
        return self._window_length

    @window_length.setter
    def window_length(self, value):
        self._window_length = value

        if self._n_fft_bins_needs_update:
            self._n_fft_bins = value

        if self._hop_length_needs_update:
            self._hop_length = value // 2

    @property
    def hop_length(self):
        """
        Number of samples that ``e_stft`` will jump ahead for every time slice.
        By default, this is equal to half of ``self.window_length`` and will update when you
        change self.window_length to stay at ``self.window_length // 2``. If you set ``self.hop_length`` directly
        then ``self.hop_length`` and ``self.window_length`` are unlinked.
        **This property is settable.**
        """
        return self._hop_length

    @hop_length.setter
    def hop_length(self, value):
        self._hop_length_needs_update = False
        self._hop_length = value

    @property
    def n_fft_bins(self):
        """
        Number of fft bins per time slice in the stft. A time slice is of length window length.
        By default the number of FFT bins is equal to window_length (value of window_length),
        but if this is set manually then e_stft takes a window of length.
        If you give it a value lower than ``self.window_length``, ``self.window_length`` will be used.
        **This property is settable.**

        """
        return self._n_fft_bins

    @n_fft_bins.setter
    def n_fft_bins(self, value):
        """
        Number of fft bins per time slice in the stft. A time slice is of length window length.
        By default the number of FFT bins is equal to window_length (value of window_length),
        but if this is set manually then e_stft takes a window of length.
        | If you give it a value lower than ``self.window_length``, ``self.window_length`` will be used.
        **This property is settable.**

        """
        # TODO: add warning for this
        if value < self.window_length:
            value = self.window_length

        self._n_fft_bins_needs_update = False
        self._n_fft_bins = value

    @property
    def window_overlap(self):
        """
        Returns number of samples of overlap between adjacent time slices.
        This is calculated like ``self.window_length - self.hop_length``.
        **This property is not settable.**
        """
        return self.window_length - self.hop_length



import os
from nussl import Duet, AudioSignal


def main():
    # Load input file
    input_file_name = '../Input/dev1_female3_inst_mix.wav'
    signal = AudioSignal(path_to_input_file=input_file_name)

    # make a directory to store output if needed
    if not os.path.exists(os.path.join('..', 'Output/')):
        os.mkdir(os.path.join('..', 'Output/'))

    # Set up DUET
    duet = Duet(signal, 3)
    # duet.stft_params.hop_length = 200
    # and run
    duet.run()

    # plot histogram results
    duet.plot('../Output/2d.png')
    duet.plot('../Output/3d.png', three_d_plot=True)

    # Create output file for each source found
    output_name_stem = '../Output/duet_source'
    i = 1
    for s in duet.make_audio_signals():
        output_file_name = output_name_stem + str(i) + '.wav'
        s.write_audio_to_file(output_file_name)
        i += 1


if __name__ == '__main__':
    main()



import nussl
import numpy as np

def main():
    win_type = nussl.WindowType.HANN
    sample_rate = nussl.DEFAULT_SAMPLE_RATE

    # Plot a simple sine wave at 5kHz
    dt = 1.0 / float(sample_rate)
    dur = 10.0 # sec
    freq = 5000 # Hz
    x = np.arange(0.0, dur, dt)
    x = np.sin(2 * np.pi * freq * x)

    signal = nussl.AudioSignal(audio_data_array=x)
    nussl.plot_stft(signal.get_channel(1), '../Output/sine_wave5000Hz.png', window_type=win_type)

    # Make a FM signal and plot it.
    lfo_freq = 3
    lfo_amp = 800
    freq2 = 15000
    x2 = np.arange(0.0, dur, dt)
    modulator = lfo_amp * np.sin(2 * np.pi * lfo_freq * x2)
    x2 = np.sin(2 * np.pi * freq2 * x2 + modulator)
    x2 += x

    signal2 = nussl.AudioSignal(audio_data_array=x2)
    nussl.plot_stft(signal2.get_channel(1), '../Output/fm_wave.png', window_type=win_type)

    # Plot from a file
    path = '../Input/police_noisy.wav'
    a = nussl.AudioSignal(path)
    nussl.plot_stft(a.get_channel(1), '../Output/police_noisy.png', freq_max=18000)

    # Plot through audio_signal
    a.plot_spectrogram()

    # plot glenn gould
    # glenn = nussl.AudioSignal('../Input/wtc-gould.mp3')
    # glenn.plot_spectrogram()


if __name__ == '__main__':
    main()


import nussl
import numpy as np
import scipy.io.wavfile as wav

def main():
    sr1, ch1 = wav.read('../Input/src1.wav')
    sr2, ch2 = wav.read('../Input/src2.wav')

    arr = np.array([ch1, ch2[0:sr1*10]], dtype=int)

    #sig2 = nussl.AudioSignal('../Input/src1.wav')
    signal = nussl.AudioSignal(audio_data_array=arr)

    n_ch = signal.num_channels
    signal.write_audio_to_file('../Output/duet_test1.wav')

if __name__ == '__main__':
    main()